'' =================================================================================================
''
''   File....... michilson_dmx.spin
''   Purpose.... 
''   Author..... Jon "JonnyMac" McPhalen
''               -- see below for terms of use
''   E-mail.....  
''   Started.... 
''   Updated.... 23 SEP 2014
''
'' =================================================================================================

{

   DMX Stream Usage (Base Address +)

   +0    - Relay 1  (0..127 = OFF, 128..255 = ON)
   +1    - Relay 2 
   +2    - Actuator position
   +3    - RGB Red
   +4    - RGB Green
   +5    - RGB Blue

   Cog Usage

   Cog 0 - Main program                                 SPIN
   Cog 1 - DMX input                                    PASM
   Cog 2 - PWM output (OUT4..OUT7)                      PASM
   Cog 3 - TTL inputs scanner/debouce                   SPIN
   Cog 4 - Actuator control                             SPIN
   Cog 5 -
   Cog 6 -
   Cog 7 -
  
} 


con { speed & timing }

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                          ' 5MHz crystal

  CLK_FREQ = ((_clkmode - xtal1) >> 6) * _xinfreq
  MS_001   = CLK_FREQ / 1_000
  US_001   = CLK_FREQ / 1_000_000
  

con { io pins }

  RX1     = 31                                                  ' programming / debug port
  TX1     = 30
  
  SDA     = 29                                                  ' eeprom
  SCL     = 28

  OUT_EN  = 27                                                  ' outputs enable pin (low)     
  
  SIO     = 26                                                  ' TTL serial io
  
  RX2     = 23                                                  ' RS-485 port
  RX2_EN  = 22
  TX2_EN  = 21
  TX2     = 20

  R_LED   = 18                                                  ' bi-color LED
  G_LED   = 17

  OUT7    = 15                                                  ' outputs
  OUT6    = 14   
  OUT5    = 13 
  OUT4    = 12
  OUT3    = 11
  OUT2    = 10   
  OUT1    =  9 
  OUT0    =  8
          
  OPT_BR  =  7                                                  ' options swithes
  OPT_A1  =  6   
  OPT_A0  =  5
  OPT_SM  =  4

  DMX_A8  =  3                                                  ' high-bit off DMX address

  DO_165  =  2                                                  ' control lines for 74x165
  CLK_165 =  1 
  LD_165  =  0


con

   #0, OFF, RED, GRN, YEL                                       ' for r/g led

   #0, NOCON, RX, TX, RXTX                                      ' for rs-485

   #0, LSBFIRST, MSBFIRST                                       ' for 74x165 shift-register                                               


con { pst formatting }

   #1, HOME, GOTOXY, #8, BKSP, TAB, LF, CLREOL, CLRDN, CR 
  #14, GOTOX, GOTOY, CLS


obj

  dmx  : "jm_dmx_rx"                                            ' DMX receiver
  outs : "jm_pwm8"                                              ' pwm driver for outputs
                                                                 
' term : "fullduplexserial"                                     ' terminal IO (debugging)
                                                                 
  
var


pub main | t, addr, check 

  setup_io                                                      ' configure HC-8+ io pins
  setup_objects                                                 ' configure software objects

  repeat
    addr := dmxaddr                                             ' read address switch
    
    ' process relay 1

    if (ttlpins & %0000_0001)                                   ' check IN0 active (override)
      low(OUT0)                                                 ' kill output    
    else
      check := dmx.read(addr+0)
      if (check => 128)                                         ' 50% or higher?
        high(OUT0)                                              ' relay on
      else
        low(OUT0)                                               ' relay off
                                                         
      
    ' process relay 2

    if (ttlpins & %0000_0001)                                   ' check IN0 active (override)
      low(OUT1)                                                 ' kill output    
    else
      check := dmx.read(addr+1)
      if (check => 128)                                         ' 50% or higher?
        high(OUT1)                                              ' relay on
      else
        low(OUT1)                                               ' relay off
        

    ' process actuator
    ' -- provides target value for actuator cog
    ' -- target is specified in 0.1s movement
      
    check := dmx.read(addr+2)                                   ' read position index
    case check
      0..63:
        acttgt := 0_0                                           ' home

      64..127:
        acttgt := 4_0                                           ' move to 4s position

      128..191:
        acttgt := 8_0                                           ' move to 8s position

      192..255:
        acttgt := 12_0                                          ' move to 12s position


    ' set rgb levels
      
    outs.set(0, addr+3)                                         ' red    OUT4
    outs.set(1, addr+4)                                         ' green  OUT5
    outs.set(2, addr+5)                                         ' blue   OUT6


pub setup_io

'' Configure for basic IO control

  outa[OUT7..OUT0] := %0000_0000                                ' clear outputs
  dira[OUT7..OUT0] := %1111_1111                                ' output mode                                  
  low(OUT_EN)                                                   ' enable output driver

  set_rs485(RX)                                                 ' allow RS-485 rx (DMX)

  set_rgled(OFF)

  
pub setup_objects | check

'' Setup objects used by the application

  dmx.start(RX2, G_LED, 1000)                                   ' start DMX (green LED = has DMX) 

  outs.start(4, OUT4)                                           ' start outputs driver

  start_scanner                                                 ' digital inputs scanner

  start_actuator                                                ' actuator movement control
  
' term.start(RX1, TX1, %0000, 115_200)                          ' start terminal (for debugging)
  
  
con

  { --------------------------------- }
  {  S U P P O R T   R O U T I N E S  } 
  { --------------------------------- }


pub pause(ms) | t

'' Delay program in milliseconds

  if (ms < 1)                                                   ' delay must be at least 1
    return
  else
    t := cnt - 1792                                             ' sync with system counter
    repeat ms                                                   ' run delay
      waitcnt(t += MS_001)


pub high(pin)

'' Makes pin output and high

  outa[pin] := 1
  dira[pin] := 1


pub low(pin)

'' Makes pin output and low

  outa[pin] := 0
  dira[pin] := 1


pub toggle(pin)

'' Toggles output pin

  !outa[pin]                                                    ' invert state
  dira[pin] := 1  
  

pub input(pin)

  dira[pin] := 0

  return ina[pin]


pub set_rgled(state)

'' Sets R/G LED
'' -- direct control only

  case state                                                    ' set color to state
    OFF : outa[R_LED..G_LED] := %00
    RED : outa[R_LED..G_LED] := %10
    GRN : outa[R_LED..G_LED] := %01
    YEL : outa[R_LED..G_LED] := %00                             ' yellow requires active driver

  dira[R_LED..G_LED] := %11                                     ' make outputs


pub set_rs485(state)

'' Sets RS-485 RX and TX enable inputs

  case state
    NOCON : outa[RX2_EN..TX2_EN] := %10
    RX    : outa[RX2_EN..TX2_EN] := %00 
    TX    : outa[RX2_EN..TX2_EN] := %11   
    RXTX  : outa[RX2_EN..TX2_EN] := %01

  dira[RX2_EN..TX2_EN] := %11


con  

  { --------------------------------------------- }
  {  D I G I T A L   I N P U T S   S C A N N E R  }
  { --------------------------------------------- }

  
var

  long  scancog
  long  scanstack[32]

  long  digins                                                  ' raw digitial inputs
  long  ttlpins                                                 ' extracted ttlpins
  long  dmxaddr                                                 ' extracted dmx address


pri start_scanner

'' Starts background scanner for IN0..IN7 pins

  stop_scanner

  scancog := cognew(digin_scanner, @scanstack) + 1              ' start scanner cog

  return scancog


pri stop_scanner

  if (scancog)                                                  ' if running
    cogstop(scancog-1)                                          '  stop
    scancog := 0                                                '  mark stopped

  longfill(@digins, 0, 3)                                       ' reset values


pri digin_scanner | t, work

'' Scans digital inputs IN0..IN7
'' -- debounce 22ms

  high(LD_165)                                                  ' setup x165 io pins
  low(CLK_165)
  input(DO_165)

  t := cnt
  repeat
    work := x165_in | (ina[DMX_A8] << 16)                       ' initial scan
    repeat 11
      waitcnt(t += constant(2 * MS_001))                        ' hold
      work &= x165_in | (ina[DMX_A8] << 16)                     ' update scan

    digins  := work
    ttlpins := digins & $FF
    dmxaddr := (digins >> 8) & $01FF

      
pri x165_in | tmp165

'' Input value from dual 74x165s on board
'' -- bit order is MSBFIRST      
                                                                                          
  outa[LD_165] := 0                                             ' blip Shift/Load line    
  outa[LD_165] := 1
         
  repeat 16                                                     ' read 16 bits         
    tmp165 := (tmp165 << 1) | ina[DO_165]                       ' get new bit             
    outa[CLK_165] := 1                                          ' blip clock              
    outa[CLK_165] := 0
   
  return tmp165 & $FFFF
  

con  

  { ----------------------------------- }
  {  A C T U A T O R   M O V E M E N T  }
  { ----------------------------------- }

  
var

  long  actcog
  long  actstack[32]

  long  acttgt                                                  ' actuator target (from DMX)
  

pri start_actuator | check

'' Starts actuator control cog
'' -- will home actuator before returning to calling code
'' -- starts scanner cog (for limit input on IN0) if not running

  stop_actuator

  ' reverse actuator to home position
  ' -- no feedback; must use maximum timing

  low(OUT2)                                                     ' reverse
  high(OUT3)
  pause(15_000)                                                 ' hold 15s
  low(OUT2)                                                     ' stop
  low(OUT3) 

  actcog := cognew(move_actuator, @actstack) + 1                ' start actuator cog

  return actcog
    

pri stop_actuator

  if (actcog)                                                   ' if running
    cogstop(actcog-1)                                           '  stop
    actcog := false                                             '  mark stopped


pri move_actuator  | actpos, t, check

'' Controls movement of actuator
'' -- moves to position specified by acttget (actuator target)
'' -- target specified in 0.1s units
'' -- monitors TTL IN0 for home position

  low(OUT2)                                                     ' stop
  low(OUT3)
  
  actpos := 0                                                   ' reset
  acttgt := 0

  ' actuator control
  ' -- moves actuator (0.1s increments) to target  

  t := cnt
  repeat                                                        ' main movement loop
    waitcnt(t += constant(100 * MS_001))                        ' position in 0.1s units
    
    if (actpos < acttgt)                                        ' extend
      high(OUT2)
      low(OUT3)
      actpos += 1                                               ' increment position    

    elseif (actpos > acttgt)                                    ' retract
      low(OUT2)                                                 ' retract                                             ' 
      high(OUT3)  
      actpos -= 1                                               ' decrement position 

    else                                                        ' at target; stop
      low(OUT2)
      low(OUT3)


dat { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}      